(*----------------------------------------------------------------------------*)
(* TJpQueue ... a queue class. Also known as FIFO (First in, first out).      *)
(* Joachim Pimiskern, 2004, Freeware.                                         *)
(*----------------------------------------------------------------------------*)
unit jpqueue;

interface

type
  pJPQueueItem = ^TJPQueueItem;
  TJpQueueItem = record
    Item: pointer;
    next: pJPQueueItem;
    prev: pJPQueueItem;
  end;

  TJpQueue = class
  private
    head: pJPQueueItem;
    tail: pJPQueueItem;
  public
    constructor Create;
    destructor  Destroy; override;
    procedure   Add(Item: pointer);
    function    Get(var obj: pointer): boolean;
    procedure   Clear;
  end;

implementation
uses
  sysutils;

var
  heap,h: pJPQueueItem;



(*----------------------------------------------------------------------------*)
(* Provide a new element container. If the assistant data structure heap is   *)
(* empty, get it from the memory heap, otherwise from the own heap.           *)
(*----------------------------------------------------------------------------*)
function newJpQueueItem: pJpQueueItem;
begin
  if (heap = nil) then
     new(result)
  else
     begin
       result := heap;
       heap := heap^.next;
     end;
end;


(*----------------------------------------------------------------------------*)
(* Give an element contaier back to the own heap. At the end of the program,  *)
(* this internal heap is freed. See the finalization section.                 *)
(*----------------------------------------------------------------------------*)
procedure freeJpQueueItem(p: pJpQueueItem);
begin
  p^.next := heap;
  heap := p;
end;



constructor TJpQueue.Create;
begin
  inherited Create;
  head := nil;
  tail := nil;
end;

destructor TJpQueue.Destroy;
begin
  inherited Destroy;
end;


procedure TJpQueue.Add(Item: pointer);
var h: pJpQueueItem;
begin
  (*--- Empty queue ---*)
  if ((head = nil) and (tail = nil)) then
     begin
       h    := newJpQueueItem();

       h^.Item := Item;
       h^.Next := nil;
       h^.Prev := nil;

       head := h;
       tail := h;
     end
  else
  if ((head <> nil) and (tail <> nil)) then
     begin
       h := newJpQueueItem();

       h^.Item := Item;
       h^.Next := tail;
       h^.Prev := nil;

       tail^.Prev := h;
       tail := h;
     end
  else
     raise Exception.Create('TJpQueue: Programming error in add()');
end;


function TJpQueue.Get(var obj: pointer): boolean;
var h: pJpQueueItem;
begin
  if (head <> nil) then
     begin
       result := true;
       obj := head^.Item;
       h := head;
       head := head^.prev;
       freeJpQueueItem(h);

       if (head = nil) then
          tail := nil;
     end
  else
     result := false;
end;

procedure TJpQueue.Clear;
var h: pointer;
begin
  while Get(h) do;
end;


initialization
  heap := nil;
finalization

  while (heap <> nil) do
  begin
    h := heap^.Next;
    dispose(heap);
    heap := h;
  end;

end.
